"""
Loop tailor is a script to loop tailor the extrusion hair of a gcode file.

The default 'Activate Loop Tailor' checkbox is on.  When it is on, the functions described below will work, when it is off, the functions
will not be called.

Loop tailor clips the ends of loops and jitters position to a different place on each layer to prevent the a ridge from forming.  The
"Clip Over Extrusion Width (ratio)" is the ratio of the amount the loops are clipped over the extrusion width.  If the ratio is too high
loops will have a gap, if the ratio is too low there will be a bulge at the loop ends.  The "Jitter Over Extrusion Width (ratio)" is the
ratio of the amount the loop ends will be jittered.  A high value means the loops will start all over the place and a low value means
loops will start at roughly the same place on each layer.  To run loop tailor, in a shell type:
> python loop_tailor.py

The following examples loop tailor the files Hollow Square.gcode & Hollow Square.gts.  The examples are run in a terminal in the folder
which contains Hollow Square.gcode, Hollow Square.gts and loop_tailor.py.  The loop_tailor function will loop_tailor if
'Activate Loop Tailor' is true, which can be set in the dialog or by changing the preferences file 'loop_tailor.csv' in the '.skeinforge' folder
in your home directory with a text editor or a spreadsheet program set to separate tabs.  The functions writeOutput and
getLoopTailorChainGcode check to see if the text has been loop_tailored, if not they call getTowerChainGcode in tower.py to tower the
text; once they have the towered text, then they loop_tailor.


> python loop_tailor.py
This brings up the dialog, after clicking 'Loop Tailor', the following is printed:
File Hollow Square.gts is being chain loop tailored.
The loop tailored file is saved as Hollow Square_loop_tailor.gcode


> python oozebane.py Hollow Square.gts
File Hollow Square.gts is being chain loop tailored.
The loop tailored file is saved as Hollow Square_loop_tailor.gcode


> python
Python 2.5.1 (r251:54863, Sep 22 2007, 01:43:31)
[GCC 4.2.1 (SUSE Linux)] on linux2
Type "help", "copyright", "credits" or "license" for more information.
>>> import loop_tailor
>>> loop_tailor.main()
This brings up the loop tailor dialog.


>>> loop_tailor.writeOutput()
Hollow Square.gts
File Hollow Square.gts is being chain loop tailored.
The loop tailored file is saved as Hollow Square_loop_tailor.gcode


>>> loop_tailor.getLoopTailorGcode("
( GCode generated by May 8, 2008 slice.py )
( Extruder Initialization )
..
many lines of gcode
..
")


>>> loop_tailor.getLoopTailorChainGcode("
( GCode generated by May 8, 2008 slice.py )
( Extruder Initialization )
..
many lines of gcode
..
")

"""

from __future__ import absolute_import
#Init has to be imported first because it has code to workaround the python bug where relative imports don't work if the module is imported as a main module.
import __init__

from skeinforge_tools.skeinforge_utilities.vec3 import Vec3
from skeinforge_tools.skeinforge_utilities import euclidean
from skeinforge_tools.skeinforge_utilities import gcodec
from skeinforge_tools.skeinforge_utilities import intercircle
from skeinforge_tools.skeinforge_utilities import preferences
from skeinforge_tools import analyze
from skeinforge_tools import import_translator
from skeinforge_tools import polyfile
from skeinforge_tools import tower
import cStringIO
import math
import sys
import time


__author__ = "Enrique Perez (perez_enrique@yahoo.com)"
__date__ = "$Date: 2008/21/04 $"
__license__ = "GPL 3.0"

def getLoopTailorChainGcode( filename, gcodeText, loopTailorPreferences = None ):
	"Loop tailor a gcode linear move text.  Chain loop tailor the gcode if it is not already loop tailored."
	gcodeText = gcodec.getGcodeFileText( filename, gcodeText )
	if not gcodec.isProcedureDone( gcodeText, 'tower' ):
		gcodeText = tower.getTowerChainGcode( filename, gcodeText )
	return getLoopTailorGcode( gcodeText, loopTailorPreferences )

def getLoopTailorGcode( gcodeText, loopTailorPreferences = None ):
	"Loop tailor a gcode linear move text."
	if gcodeText == '':
		return ''
	if gcodec.isProcedureDone( gcodeText, 'loop_tailor' ):
		return gcodeText
	if loopTailorPreferences == None:
		loopTailorPreferences = LoopTailorPreferences()
		preferences.readPreferences( loopTailorPreferences )
	if not loopTailorPreferences.activateLoopTailor.value:
		return gcodeText
	skein = LoopTailorSkein()
	skein.parseGcode( gcodeText, loopTailorPreferences )
	return skein.output.getvalue()

def writeOutput( filename = '' ):
	"Loop tailor a gcode linear move file.  Chain loop tailor the gcode if it is not already loop tailored.  If no filename is specified, loop tailor the first unmodified gcode file in this folder."
	if filename == '':
		unmodified = import_translator.getGNUTranslatorFilesUnmodified()
		if len( unmodified ) == 0:
			print( "There are no unmodified gcode files in this folder." )
			return
		filename = unmodified[ 0 ]
	loopTailorPreferences = LoopTailorPreferences()
	preferences.readPreferences( loopTailorPreferences )
	startTime = time.time()
	print( 'File ' + gcodec.getSummarizedFilename( filename ) + ' is being chain loop tailored.' )
	suffixFilename = filename[ : filename.rfind( '.' ) ] + '_loop_tailor.gcode'
	loopTailorGcode = getLoopTailorChainGcode( filename, '', loopTailorPreferences )
	if loopTailorGcode == '':
		return
	gcodec.writeFileText( suffixFilename, loopTailorGcode )
	print( 'The loop tailored file is saved as ' + gcodec.getSummarizedFilename( suffixFilename ) )
	analyze.writeOutput( suffixFilename, loopTailorGcode )
	print( 'It took ' + str( int( round( time.time() - startTime ) ) ) + ' seconds to loop tailor the file.' )


class LoopTailorSkein:
	"A class to loop tailor a skein of extrusions."
	def __init__( self ):
		self.decimalPlacesCarried = 3
		self.feedrateMinute = None
		self.isLoopPerimeter = False
		self.layerGolden = 0.0
		self.loopPath = None
		self.lineIndex = 0
		self.lines = None
		self.oldLocation = None
		self.output = cStringIO.StringIO()

	def addGcodeFromThread( self, thread ):
		"Add a gcode thread to the output."
		if len( thread ) > 0:
			self.addGcodeMovement( thread[ 0 ] )
		else:
			print( "zero length vertex positions array which was skipped over, this should never happen" )
		if len( thread ) < 2:
			return
		self.addLine( 'M101' )
		for point in thread[ 1 : ]:
			self.addGcodeMovement( point )
#		self.addLine( "M103" ) # Turn extruder off.

	def addGcodeMovement( self, point ):
		"Add a movement to the output."
		if self.feedrateMinute == None:
			self.addLine( "G1 X%s Y%s Z%s" % ( self.getRounded( point.x ), self.getRounded( point.y ), self.getRounded( point.z ) ) )
		else:
			self.addLine( "G1 X%s Y%s Z%s F%s" % ( self.getRounded( point.x ), self.getRounded( point.y ), self.getRounded( point.z ), self.getRounded( self.feedrateMinute ) ) )

	def addLine( self, line ):
		"Add a line of text and a newline to the output."
		self.output.write( line + "\n" )

	def addTailoredLoopPath( self ):
		"Add a clipped and jittered loop path."
		if self.layerJitter == 0.0 and self.clip <= 0.0:
			self.addGcodeFromThread( self.loopPath )
			self.loopPath = None
		jitteredLoopPath = self.getJitteredLoopPath( self.loopPath )
		clippedLoopPath = self.getClippedLoopPath( jitteredLoopPath )
		clippedLoopPath = euclidean.getSimplifiedPath( clippedLoopPath, self.extrusionWidth )
		self.addGcodeFromThread( clippedLoopPath )
		self.loopPath = None

	def getClippedLoopPath( self, loopPath ):
		"Get a clipped loop path."
		if self.clip <= 0.0:
			return loopPath
		loopPathLength = euclidean.getPathLength( loopPath )
		lastLength = 0.0
		pointIndex = 0
		totalLength = 0.0
		clippedLength = loopPathLength - self.clip
		while totalLength < clippedLength and pointIndex < len( loopPath ) - 1:
			firstPoint = loopPath[ pointIndex ]
			secondPoint  = loopPath[ pointIndex + 1 ]
			pointIndex += 1
			lastLength = totalLength
			totalLength += firstPoint.distance( secondPoint )
		remainingLength = clippedLength - lastLength
		clippedLoopPath = loopPath[ : pointIndex ]
		ultimateClippedPoint = loopPath[ pointIndex ]
		penultimateClippedPoint = clippedLoopPath[ - 1 ]
		segment = ultimateClippedPoint.minus( penultimateClippedPoint )
		segmentLength = segment.length()
		if segmentLength <= 0.0:
			return clippedLoopPath
		newUltimatePoint = penultimateClippedPoint.plus( segment.times( remainingLength / segmentLength ) )
		return clippedLoopPath + [ newUltimatePoint ]

	def getJitteredLoopPath( self, loopPath ):
		"Get a jittered loop path."
		loopPathLength = euclidean.getPathLength( loopPath )
		lastLength = 0.0
		pointIndex = 0
		totalLength = 0.0
		jitterPosition = ( self.layerJitter + 256.0 * loopPathLength ) % loopPathLength
		jitterLoop = loopPath[ : - 1 ]
		while totalLength < jitterPosition and pointIndex < len( jitterLoop ):
			firstPoint = jitterLoop[ pointIndex ]
			secondPoint  = jitterLoop[ ( pointIndex + 1 ) % len( jitterLoop ) ]
			pointIndex += 1
			lastLength = totalLength
			totalLength += firstPoint.distance( secondPoint )
		remainingLength = jitterPosition - lastLength
		pointIndex = pointIndex % len( jitterLoop )
		ultimateJitteredPoint = jitterLoop[ pointIndex ]
		penultimateJitteredPointIndex = ( pointIndex + len( jitterLoop ) - 1 ) % len( jitterLoop )
		penultimateJitteredPoint = jitterLoop[ penultimateJitteredPointIndex ]
		segment = ultimateJitteredPoint.minus( penultimateJitteredPoint )
		segmentLength = segment.length()
		originalOffsetLoop = euclidean.getAroundLoop( pointIndex, pointIndex, jitterLoop )
		if segmentLength <= 0.0:
			return [ penultimateJitteredPoint ] + originalOffsetLoop
		newUltimatePoint = penultimateJitteredPoint.plus( segment.times( remainingLength / segmentLength ) )
		return [ newUltimatePoint ] + originalOffsetLoop + [ newUltimatePoint ]

	def getRounded( self, number ):
		"Get number rounded to the number of carried decimal places as a string."
		return euclidean.getRoundedToDecimalPlaces( self.decimalPlacesCarried, number )

	def isNextExtruderOn( self ):
		"Determine if there is an extruder on command before a move command."
		line = self.lines[ self.lineIndex ]
		splitLine = line.split()
		for afterIndex in xrange( self.lineIndex + 1, len( self.lines ) ):
			line = self.lines[ afterIndex ]
			splitLine = line.split()
			firstWord = gcodec.getFirstWord( splitLine )
			if firstWord == 'G1' or firstWord == 'M103':
				return False
			elif firstWord == 'M101':
				return True
		return False

	def linearMove( self, splitLine ):
		"Add travel move around loop paths if this the extruder is off."
		location = gcodec.getLocationFromSplitLine( self.oldLocation, splitLine )
		self.feedrateMinute = gcodec.getFeedrateMinute( self.feedrateMinute, splitLine )
		if self.isLoopPerimeter:
			if self.isNextExtruderOn():
				self.loopPath = []
		if self.loopPath != None:
			self.loopPath.append( location )
		self.oldLocation = location

	def parseGcode( self, gcodeText, loopTailorPreferences ):
		"Parse gcode text and store the loop tailor gcode."
		self.lines = gcodec.getTextLines( gcodeText )
		self.parseInitialization( loopTailorPreferences )
		for self.lineIndex in xrange( self.lineIndex, len( self.lines ) ):
			line = self.lines[ self.lineIndex ]
			self.parseLine( line )

	def parseInitialization( self, loopTailorPreferences ):
		"Parse gcode initialization and store the parameters."
		for self.lineIndex in xrange( len( self.lines ) ):
			line = self.lines[ self.lineIndex ]
			splitLine = line.split()
			firstWord = gcodec.getFirstWord( splitLine )
			if firstWord == '(<extrusionWidth>':
				self.extrusionWidth = float( splitLine[ 1 ] )
				self.clip = loopTailorPreferences.clipOverExtrusionWidth.value * self.extrusionWidth
				self.jitter = loopTailorPreferences.jitterOverExtrusionWidth.value * self.extrusionWidth
			elif firstWord == '(<decimalPlacesCarried>':
				self.decimalPlacesCarried = int( splitLine[ 1 ] )
			elif firstWord == '(<extrusionStart>':
				self.addLine( '(<procedureDone> loop_tailor )' )
				return
			self.addLine( line )

	def parseLine( self, line ):
		"Parse a gcode line and add it to the loop tailor skein."
		splitLine = line.split()
		if len( splitLine ) < 1:
			return
		firstWord = splitLine[ 0 ]
		if firstWord == 'G1':
			self.linearMove( splitLine )
		elif firstWord == 'M103':
			self.isLoopPerimeter = False
			if self.loopPath != None:
				self.addTailoredLoopPath()
		elif firstWord == '(<layerStart>':
			self.layerGolden += 0.61803398874989479
			self.layerJitter = self.jitter * ( math.fmod( self.layerGolden, 1.0 ) - 0.5 ) + 0.5 * self.clip
		if firstWord == '(<loop>' or firstWord == '(<perimeter>':
			self.isLoopPerimeter = True
		if self.loopPath == None:
			self.addLine( line )


class LoopTailorPreferences:
	"A class to handle the loop tailor preferences."
	def __init__( self ):
		"Set the default preferences, execute title & preferences filename."
		#Set the default preferences.
		self.archive = []
		self.activateLoopTailor = preferences.BooleanPreference().getFromValue( 'Activate Loop Tailor', True )
		self.archive.append( self.activateLoopTailor )
		self.clipOverExtrusionWidth = preferences.FloatPreference().getFromValue( 'Clip Over Extrusion Width (ratio):', 0.3 )
		self.archive.append( self.clipOverExtrusionWidth )
		self.filenameInput = preferences.Filename().getFromFilename( import_translator.getGNUTranslatorGcodeFileTypeTuples(), 'Open File to be Loop Tailored', '' )
		self.archive.append( self.filenameInput )
		self.jitterOverExtrusionWidth = preferences.FloatPreference().getFromValue( 'Jitter Over Extrusion Width (ratio):', 2.0 )
		self.archive.append( self.jitterOverExtrusionWidth )
		#Create the archive, title of the execute button, title of the dialog & preferences filename.
		self.executeTitle = 'Loop Tailor'
		self.filenamePreferences = preferences.getPreferencesFilePath( 'loop_tailor.csv' )
		self.filenameHelp = 'skeinforge_tools.loop_tailor.html'
		self.saveTitle = 'Save Preferences'
		self.title = 'Loop Tailor Preferences'

	def execute( self ):
		"Loop tailor button has been clicked."
		filenames = polyfile.getFileOrDirectoryTypesUnmodifiedGcode( self.filenameInput.value, import_translator.getGNUTranslatorFileTypes(), self.filenameInput.wasCancelled )
		for filename in filenames:
			writeOutput( filename )


def main( hashtable = None ):
	"Display the loop tailor dialog."
	if len( sys.argv ) > 1:
		writeOutput( ' '.join( sys.argv[ 1 : ] ) )
	else:
		preferences.displayDialog( LoopTailorPreferences() )

if __name__ == "__main__":
	main()
